;; Generated package description from org.el  -*- no-byte-compile: t -*-
(define-package "org" "9.2.6" "Outline-based notes management and organizer" 'nil :keywords '("outlines" "hypermedia" "calendar" "wp") :authors '(("Carsten Dominik <carsten at orgmode dot org>")) :maintainer '("Carsten Dominik <carsten at orgmode dot org>") :url "https://orgmode.org")
